import serial
import socket

from .types import SerialOrNetwork
from .comm import hanazeder_encode_msg, hanazeder_read, hanazeder_decode_num


class ConnectError(Exception):
    pass

class NotConnectedError(Exception):
    pass

class ConnectionInvalidError(Exception):
    pass

class HanazederFP:
    HEADER = b'\xEE'
    last_msg_num = 1
    connected = False
    connection: SerialOrNetwork

    def __init__(self,
        serial_port="/dev/ttyUSB0",
        address=None,
        port=None,
        timeout=1):
        if serial_port:
            self.connection = serial.Serial(
                port=serial_port,
                baudrate = 38400,
                parity=serial.PARITY_NONE,
                stopbits=serial.STOPBITS_ONE,
                bytesize=serial.EIGHTBITS,
                timeout=timeout
            )
        elif address and port:
            self.connection = socket.create_connection((address, port), timeout).makefile()
        else:
            raise ConnectionInvalidError("Specify either address and port or serial port")
    
    def connect(self) -> bool:
        handshake = hanazeder_encode_msg(self.HEADER, 0, b'\x01\x00')
        self.connection.write(handshake)
        response = self.connection.read(size=10)
        if response is not b'\xee\x00\xf0\x05\x00\x00\x02\x01\x06\xf9':
            raise ConnectError(f'Wrong response for handshake, got {response}')
        self.connected = True
        return self.connected
    
    def create_read_register_msg(self, register: int) -> bytes:
        request = bytes(b'\x04\x01') + bytes(register)
        return hanazeder_encode_msg(self.HEADER, self.last_msg_num, request)
    
    def read_register(self, register: int) -> float:
        if not self.connected:
            raise NotConnectedError()
        self.last_msg_num = (self.last_msg_num + 1) % 256
        msg = self.create_read_register_msg(register)
        self.connection.write(msg)
        response = hanazeder_read(self.HEADER, self.last_msg_num, self.connection)
        value = hanazeder_decode_num(self.HEADER, response)
        return value
        



from .Hanazeder import HanazederFP
import argparse
import sys

def main() -> int:
    parser = argparse.ArgumentParser()
    parser.add_argument("--serial-port", help="set serial port",
                    type=str)
    parser.add_argument("--register", help="register to read",
                    type=int)
    parser.add_argument("--address", help="connect to HOSTNAME, needs port as well",
                    type=str)
    parser.add_argument("--port", help="connect to HOSTNAME on port PORT",
                    type=int, default=5000)
    args = parser.parse_args()

    if args.address and args.serial_port:
        print('Cannot specify both serial-port and address')
        return 1
    
    if args.address and not args.port:
        print('Specify port together with address')
        return 2

    conn = HanazederFP(serial_port=args.serial_port, address=args.address, port=args.port)
    conn.connect()
    value = conn.read_register(parser.register)
    print(f'Register {args.register} is {value}')
    return 0

if __name__ == '__main__':
    sys.exit(main())


from .types import SerialOrNetwork
from .encoding import *

from serial import Serial
from crccheck.crc import Crc8Maxim

SENSOR_GONE = b'\xFF\x7F'

class ReadtimeoutException(Exception):
    pass

class ChecksumNotMatchingException(Exception):
    pass

class InvalidHeaderException(Exception):
    pass

class IllegalArgumentException(Exception):
    pass

def hanazeder_decode_num(header, value) -> float:
    unescaped = value.replace(header + header, header)
    if unescaped == SENSOR_GONE:
        return None
    int_val = int.from_bytes(value, byteorder='little', signed=True)
    return int_val / 10

def hanazeder_read(expected_header: bytes, msg_num: int, connection: SerialOrNetwork) -> bytes:
    # start by reading at least four bytes
    header = connection.read(4)
    if len(header) < 4:
        raise ReadtimeoutException("Could not read header")
    if header[0] is not expected_header[0]:
        raise InvalidHeaderException(f'Expecting header {expected_header[0]}, got {header[0]} instead')
    if header[1] != msg_num:
        raise InvalidHeaderException(f'Expecting message number {msg_num} but got {header[1]}')
    value_size = header[3]
    value = connection.read(value_size)
    # Always followed by one byte checksum
    checksum = connection.read()
    crc = Crc8Maxim()
    crc.process(header[1:])
    crc.process(value)
    calculated_crc = crc.finalbytes()
    if calculated_crc != checksum:
        raise ChecksumNotMatchingException(f'Expected checksum {calculated_crc} but got {checksum}')
    return value

def hanazeder_encode_msg(header: bytes, msg_num: int, request: bytes) -> bytes:
    if msg_num < 0 or msg_num > 255:
        raise IllegalArgumentException("msg_num must be between 0 and 255")
    if len(header) != 1:
        raise IllegalArgumentException("header must be single byte")
    msg = bytearray(header)
    msg.append(msg_num)
    msg += bytearray(request)
    checksum = Crc8Maxim.calcbytes(msg[1:])
    msg += bytearray(checksum)
    return bytes(msg)